# Copyright (c) 2013, University of Bristol and University of Southampton.

from EStat.Templating import Template

class Template2LevelMVMixedResponseRSCC(Template):
    'Fits 2-level multivariate mixed response models (Normal, binomial, ordered, or unordered), allowing for random slopes, missing data, and a choice of uniform or Wishart priors. Creates custom C code.'

    __version__ = '1.0.0'

    tags = [ 'Model', '2-Level', 'Normal', 'Multivariate response', 'Random slopes']
    engines = ['CustomC']

    inputs = '''
y = DataMatrix('Responses: ')
L2ID = IDVector('Level 2 ID: ')

for i in range(0, len(y)):
    context['bin'+str(i+1)] = Text('Response Type for response ' + y[i] + ': ', ['Normal', 'Binomial', 'Ordered', 'Unordered'])
    if(context['bin'+str(i+1)] == 'Ordered'):
       context['ncats'+y[i]] = Integer('Number of categories for ' + y[i] + ': ')
    if(context['bin'+str(i+1)] == 'Unordered'):
       context['ncats'+y[i]] = Integer('Number of categories for ' + y[i] + ': ')
       for j in range(0, context['ncats'+y[i]]-1):
          context['xf'+str(i+1)+'_'+str(j)] = DataMatrix('Explanatory variables for category ' + str(j) + ': ', allow_cat = True)
          context['xr'+str(i+1)+'_'+str(j)] = DataMatrix('Explanatory variables random for category ' + str(j) + '?: ', allow_cat = True)
    else:
       context['xf'+str(i+1)] = DataMatrix('Explanatory variables for response ' + y[i] + ': ', allow_cat = True)
       context['xr'+str(i+1)] = DataMatrix('Explanatory variables random for response ' + y[i]+ '?: ', allow_cat = True)

storeresid = Boolean('Store residuals?')
randu = Boolean('Randomise starting value for level 2 residuals?') 

lenbeta = 0
lenu = 0
n = 0
uind = 0
for i in range(0, len(y)):
    if(context['bin'+str(i+1)] == 'Normal'):
       context['miss'+y[i]] = ParamVector(monitor=False)
    if(context['bin'+str(i+1)] == 'Binomial'):
       context['misb'+y[i]] = ParamVector(monitor=False)
    if(context['bin'+str(i+1)] == 'Ordered'):
       context['miso'+y[i]] = ParamVector(monitor=False)
       context['alpha'+y[i]] = ParamVector()
       context['alpha'+y[i]].ncols = -(context['ncats'+y[i]] -1)
       context['alpha'+y[i]].set_value( map(str, range(0, (context['ncats'+y[i]] -1))) )
    if(context['bin'+str(i+1)] == 'Unordered'):
       for j in range(0, context['ncats'+y[i]]-1):
          lenbeta += len(context['xf'+str(i+1)+'_'+str(j)])
          context['misu'+str(context['ncats'+y[i]])+'_'+y[i]+'_'+ str(j)] = ParamVector(monitor=False)
          lenu += len(context['xr'+str(i+1)+'_'+str(j)])
          for x in range(0, len(context['xr'+str(i+1)+'_'+str(j)])):
              if storeresid:
                  context['u' + str(x) + '_' + str(uind)] = ParamVector(parents=[L2ID], as_scalar=False)
              else:
                  context['u' + str(x) + '_' + str(uind)] = ParamVector(parents=[L2ID], as_scalar=False, monitor=False)
          uind += 1
       context['y'+ str(i+1)] = DataMatrix(value=[])
       context['origv'+str(i+1)] = Text(value = [])
       n += context['ncats'+y[i]] - 1
    else:
       lenbeta += len(context['xf'+str(i+1)])
       n += 1
       lenu += len(context['xr'+str(i+1)])
       for x in range(0, len(context['xr'+str(i+1)])):
           if storeresid:
               context['u' + str(x) + '_' + str(uind)] = ParamVector(parents=[L2ID], as_scalar=False)
           else:
               context['u' + str(x) + '_' + str(uind)] = ParamVector(parents=[L2ID], as_scalar=False, monitor=False)
       uind += 1

if n == 1:
    tau = ParamScalar()
    sigma = ParamScalar(modelled = False)
    tau_u = ParamScalar()
    sigma_u = ParamScalar(modelled = False)
else:
    omega_e = ParamVector()
    omega_e.ncols = ((n+1)*n)/2
    omega_e.select('Starting values for Level 1 variance matrix (lower triangle): ')
    omega_u = ParamMatrix(modelled = False)
    omega_u.size = lenu
    d_e = ParamMatrix(modelled = False)
    d_e.size = n
    d_u = ParamMatrix()
    d_u.size = lenu
    priors2 = Text('Level 2 Priors: ', ['Uniform', 'Wishart'])
    if priors2 == 'Wishart':
        R2 = List('R matrix: ')
        v2 = Integer('Degrees of Freedom:')
beta = ParamVector()
beta.ncols = lenbeta
deviance = ParamScalar(modelled = False)
'''

    customccode = '''
<%!
    def mmult2(names, var, index,count):
        out = ""
        first = True
        for name in names:
            if first == False:
                out += ' + '
            else:
                first = False
            out += 'double(' + name + '[' + index + ']) * ' + var + '_' + str(count)
            count += 1
        return out

    def PyToC(var):
        if isinstance(var, list):
            return "{" + ",".join(map(PyToC, var)) + "}"
        else:
            return str(var)    
%>
    {
<% nvar = len(context['y']) %>\\
<% n = 0 %> \\
% for i in range(0,nvar):
% if (context['bin'+str(i+1)] == 'Unordered'):
<% n += context['ncats'+y[i]]-1 %> \\
% else:
<% n += 1 %> \\
% endif
% endfor

% for var in range(0,nvar):
% if (context['bin'+str(var+1)] == 'Ordered'):
    const int ncats${y[var]} = ${context['ncats'+y[var]]};
% endif  
% if (context['bin'+str(var+1)] == 'Unordered'):
    const int ncats${y[var]} = ${context['ncats'+y[var]]};
% endif  
% endfor
        // Bill's attempt to include ordered cutpoints to go in here (note y step done elsewhere)
% for i in range(0, nvar):
% if (context['bin'+str(i+1)] == 'Ordered'):
        {
            double low[] = ${PyToC([1000] * int(context['ncats'+y[i]]))};
            double upp[] = ${PyToC([-1000] * int(context['ncats'+y[i]]))};

            for(int j = 0; j < length(${y[i]}); j++) {
                if(${y[i]}[j] <= -9.999e29) {
                } else {            
                    if(miso${y[i]}[j] < low[int(${y[i]}[j])])
                        low[int(${y[i]}[j])] = miso${y[i]}[j];
                    if(miso${y[i]}[j] > upp[int(${y[i]}[j])])
                        upp[int(${y[i]}[j])] = miso${y[i]}[j];
                }
            } 
            for(int j = 1; j<ncats${y[i]}-1; j++) {
                double ul,ll;            
                if(upp[j]>double(alpha${y[i]}[j-1]))
                    ll = upp[j];
                else 
                    ll = double(alpha${y[i]}[j-1]);
                if(j == (ncats${y[i]}-2))
                    ul = low[j+1];
                else if (low[j+1] < double(alpha${y[i]}[j+1]))
                    ul = low[j+1];
                else
                    ul = double(alpha${y[i]}[j+1]);
                alpha${y[i]}[j]= ll+(ul-ll)*random01();
            }
        }
% endif
% endfor
        // Bill's attempt of step for d_e
        SymMatrix va(${n});
        double llcurr;
<% count = 0 %>\\
% for i in range(0, n):
% for j in range(0, i + 1):
        va(${i}, ${j}) = omega_e[${count}];
<% count += 1 %>\\
% endfor
% endfor
        bool fail = false;
        mat_d_e = matrix_sym_inverse(va, fail);
 // Note currently using a uniform prior for variance matrix
        // Note sb contains all terms needed in MH steps
        SymMatrix sb(${n});
        for(int i = 0; i < length(${y[0]}); i++) {
<% lenbeta = 0 %>\\
<% row = 0 %>\\
<% uind = 0 %>\\
% for i in range(0, nvar):
% if (context['bin'+str(i+1)] == 'Normal'):
            double e${row} = double(miss${y[i]}[i]) - (${mmult2(context['xf' + str(i+1)], 'beta', 'i', lenbeta)}) \\
% for j in range(0, len(context['xr'+str(i+1)])):
 - double(u${j}_${uind}[int(${L2ID}[i])]) * double(${context['xr' + str(i+1)][j]}[i])\\
%endfor
<% uind += 1 %>\\
;
<% row += 1 %>\\
% endif
% if (context['bin'+str(i+1)] == 'Binomial'):
            double e${row} = double(misb${y[i]}[i]) - (${mmult2(context['xf' + str(i+1)], 'beta', 'i', lenbeta)}) \\
% for j in range(0, len(context['xr'+str(i+1)])):
 - double(u${j}_${uind}[int(${L2ID}[i])]) * double(${context['xr' + str(i+1)][j]}[i])\\
%endfor
;
<% uind += 1 %>\\
<% row += 1 %>\\
% endif
% if (context['bin'+str(i+1)] == 'Ordered'):
            double e${row} = double(miso${y[i]}[i]) - (${mmult2(context['xf' + str(i+1)], 'beta', 'i', lenbeta)}) \\
% for j in range(0, len(context['xr'+str(i+1)])):
 - double(u${j}_${uind}[int(${L2ID}[i])]) * double(${context['xr' + str(i+1)][j]}[i])\\
%endfor
<% uind += 1 %>\\
;
<% row += 1 %>\\
% endif
% if (context['bin'+str(i+1)] == 'Unordered'):
% for k in range(0,context['ncats'+y[i]]-1):
            double e${row} = double(misu${context['ncats'+y[i]]}_${y[i]}_${k}[i]) - (${mmult2(context['xf' + str(i+1)+'_'+str(k)], 'beta', 'i', lenbeta)}) \\
% for j in range(0, len(context['xr'+str(i+1)+'_'+str(k)])):
 - double(u${j}_${uind}[int(${L2ID}[i])]) * double(${context['xr' + str(i+1)+'_'+str(k)][j]}[i])\\
%endfor
<% uind += 1 %>\\
;
<% row += 1 %>\\
<% lenbeta += len(context['xf' + str(i + 1) + '_'+ str(j)]) %>\\
% endfor
%else :
<% lenbeta += len(context['xf' + str(i + 1)]) %>\\
%endif
% endfor
% for i in range(0, n):
% for j in range(i, n):
            sb(${i}, ${j}) += e${i} * e${j};
% endfor
% endfor
        }
        // Now loop over elements of va (omega_e)
<% count = 0 %>\\
<% var1 = 0 %>\\
<% withinvar1 = 0 %>\\
% for i in range(0, n):
<% var2 = var1 %>\\
<% withinvar2 = withinvar1 %>\\
% for j in range(i, n):
        {
% if var1 == var2 and (context['bin'+str(var1+1)] != 'Normal'):
            omega_e_accept[${count}] = 44;
% else:
            llcurr = length(${y[0]})*log(matrix_determinant(mat_d_e))/2;
            for(int i1 = 0; i1 < ${n}; i1++)
                for(int j1 = 0; j1 < ${n}; j1++)
                    llcurr -= sb(i1, j1) * mat_d_e(i1, j1)/2;
            double old = va(${i}, ${j});
            va(${i}, ${j}) = dnorm(va(${i}, ${j}), 1.0/omega_e_sd[${count}]);
            mat_d_e = matrix_sym_inverse(va, fail);
            if (fail == true) {
                va(${i}, ${j}) = old;
                mat_d_e = matrix_sym_inverse(va, fail);
            } else {
                double llnew = length(${y[0]})*log(matrix_determinant(mat_d_e))/2;
                for(int i1=0;i1<${n};i1++)
                    for(int j1=0;j1<${n};j1++)
                        llnew -= sb(i1, j1) * mat_d_e(i1, j1)/2;
                int change = 1;
                if(isFinite(llnew) == 0)
                    change = 0;
                else if (llnew > llcurr)
                    ;
                else {
                    if(random01() > exp(llnew-llcurr))
                        change = 0;
                }
                if(change == 0) {
                    va(${i}, ${j}) = old;
                    mat_d_e = matrix_sym_inverse(va, fail);
                } else
                    omega_e_accept[${count}]++;
            }
% endif
<% count += 1 %>\\
        }
% if (context['bin'+str(var2+1)] != 'Unordered'):
<% var2+= 1 %>\\
% elif withinvar2 != (context['ncats'+y[var2]]-2):
<% withinvar2 += 1 %>\\
%else:
<% var2+= 1 %>\\
<% withinvar2 = 0 %>\\
% endif
% endfor
% if (context['bin'+str(var1+1)] != 'Unordered'):
<% var1+= 1 %>\\
% elif withinvar1 != (context['ncats'+y[var1]]-2):
<% withinvar1 += 1 %>\\
% else:
<% var1+= 1 %>\\
<% withinvar1 = 0 %>\\
% endif
% endfor
<% count = 0 %>\\
% for i in range(0, n):
% for j in range(0, i + 1):
                omega_e[${count}] = va(${i}, ${j});
<% count += 1 %>\\
% endfor
% endfor





<% lenu = 0 %>\\
% for i in range(0, len(y)):
% if (context['bin'+str(i+1)] == 'Unordered'):
% for j in range(0,context['ncats'+y[i]]-1):
<% lenu += len(context['xr' + str(i+1) + '_' + str(j)]) %>\\
% endfor
% else:
<% lenu += len(context['xr' + str(i+1)]) %>\\
% endif
% endfor
        SymMatrix sbu(${lenu});
        for(int i = 0; i < length(u0_0); i++) {
<% loop1 = 0 %>\\
<% uind1 = 0 %>\\
% for i1 in range(0, len(y)):
% if (context['bin'+str(i1+1)] == 'Unordered'):
% for j in range(0,context['ncats'+y[i1]]-1):
% for j1 in range(0, len(context['xr' + str(i1+1) + '_' + str(j)])):
<% loop2 = 0 %>\\
<% uind2 = 0 %>\\
% for i2 in range(0, len(y)):
% if (context['bin'+str(i2+1)] == 'Unordered'):
% for m in range(0,context['ncats'+y[i2]]-1):
% for j2 in range(0, len(context['xr' + str(i2+1) + '_' + str(m)])):
% if loop2 >= loop1:
            sbu(${loop1}, ${loop2}) += double(u${j1}_${uind1}[i]) * double(u${j2}_${uind2}[i]);
% endif
<% loop2 += 1 %>\\
% endfor
<% uind2 += 1 %>\\
% endfor
% else:
% for j2 in range(0, len(context['xr' + str(i2+1)])):
% if loop2 >= loop1:
            sbu(${loop1}, ${loop2}) += double(u${j1}_${uind1}[i]) * double(u${j2}_${uind2}[i]);
% endif
<% loop2 += 1 %>\\
% endfor
<% uind2 += 1 %>\\
% endif
% endfor
<% loop1 += 1 %>\\
% endfor
<% uind1 += 1 %>\\
% endfor
% else:
% for j1 in range(0, len(context['xr' + str(i1+1)])):
<% loop2 = 0 %>\\
<% uind2 = 0 %>\\
% for i2 in range(0, len(y)):
% if (context['bin'+str(i2+1)] == 'Unordered'):
% for m in range(0,context['ncats'+y[i2]]-1):
% for j2 in range(0, len(context['xr' + str(i2+1) + '_' + str(m)])):
%if loop2 >= loop1:
            sbu(${loop1}, ${loop2}) += double(u${j1}_${uind1}[i]) * double(u${j2}_${uind2}[i]);
% endif
<% loop2 += 1 %>\\
% endfor
<% uind2 += 1 %>\\
% endfor
% else:
% for j2 in range(0, len(context['xr' + str(i2+1)])):
%if loop2 >= loop1:
            sbu(${loop1}, ${loop2}) += double(u${j1}_${uind1}[i]) * double(u${j2}_${uind2}[i]);
% endif
<% loop2 += 1 %>\\
% endfor
<% uind2 += 1 %>\\
% endif
% endfor
<% loop1 += 1 %>\\
% endfor
<% uind1 += 1 %>\\
% endif
% endfor
        }
% if context['priors2'] == 'Uniform':
        if (runstate == 0) {
% for i in range(0, loop1):
            sbu(${i}, ${i}) += 0.0001;
% endfor
        }
% endif
% if context['priors2'] == 'Wishart':
<%
import numpy
Rmat = numpy.empty([loop1, loop1])
count = 0 
for i in range(0, loop1):
    for j in range(0, i + 1):
        Rmat[i, j] = float(context['R2'][count]) 
        Rmat[j, i] = Rmat[i, j]
        count += 1
%>

<% count = 0 %>
% for i in range(0, loop1):
% for j in range(i, loop1):
        sbu(${i}, ${j}) += ${str(Rmat[i][j] * float(context['v2']))};
% endfor
% endfor
% endif


        matrix_sym_invinplace(sbu);
% if context['priors2'] == 'Uniform':
        int vwu = length(u0_0) - ${loop1 + 1};
% endif
% if context['priors2'] == 'Wishart':
        int vwu = length(u0_0) + ${context['v2']};
% endif

        mat_d_u = dwishart(vwu, sbu, fail); 
        mat_omega_u = matrix_sym_inverse(mat_d_u, fail);



                // Step for beta
                std::vector<int> betaind(${lenbeta});
<% betacount = 0 %>
<% count = 0 %>
% for i in range(0, len(y)):
% if (context['bin'+str(i+1)] == 'Unordered'):
% for m in range(0,context['ncats'+y[i]]-1):
% for j in range(0, len(context['xf' +str(i+1)+'_'+str(m)])):
                betaind[${betacount}] = ${count};
<% betacount += 1 %>
% endfor
<% count += 1 %>
% endfor
% else:
% for j in range(0, len(context['xf' +str(i+1)])):
                betaind[${betacount}] = ${count};
<% betacount += 1 %>
% endfor
<% count += 1 %>
% endif

% endfor
                std::vector<double*> tmp_x;
% for i in range(0, len(y)):

% if (context['bin'+str(i+1)] == 'Unordered'):
% for m in range(0,context['ncats'+y[i]]-1):
% for j in range(0, len(context['xf' + str(i+1)+'_'+str(m)])):
                tmp_x.push_back((double*)${context['xf' + str(i+1)+'_' + str(m)][j]});
% endfor
% endfor
% else:
% for j in range(0, len(context['xf' + str(i+1)])):
                tmp_x.push_back((double*)${context['xf' + str(i+1)][j]});
% endfor
% endif
% endfor
                RectMatrix mat_x(tmp_x, length(${y[0]}));


                std::vector<double*> tmp_xr;
% for i in range(0, len(y)):

% if (context['bin'+str(i+1)] == 'Unordered'):
% for m in range(0,context['ncats'+y[i]]-1):
% for j in range(0, len(context['xr' + str(i+1)+'_'+str(m)])):
                tmp_xr.push_back((double*)${context['xr' + str(i+1) + '_' + str(m)][j]});
% endfor
% endfor
% else:
% for j in range(0, len(context['xr' + str(i+1)])):
                tmp_xr.push_back((double*)${context['xr' + str(i+1)][j]});
% endfor
% endif

% endfor
                RectMatrix mat_xr(tmp_xr, length(${y[0]}));

                std::vector<double*> tmp_ymiss;
<%
pyyind = []
pyymap = []
pypos = 0
%>
% for i in range(0, len(y)):
% if (context['bin'+str(i+1)] == 'Normal'):
                tmp_ymiss.push_back(miss${y[i]});
<%
pyymap.append(pypos)
pypos += 1
pyyind.append(i)
%>
% endif
% if (context['bin'+str(i+1)] == 'Binomial'):
                tmp_ymiss.push_back(misb${y[i]});
<%
pyymap.append(pypos)
pypos += 1
pyyind.append(i)
%>
% endif
% if (context['bin'+str(i+1)] == 'Ordered'):
                tmp_ymiss.push_back(miso${y[i]});
<%
pyymap.append(pypos)
pypos += 1
pyyind.append(i)
%>
% endif
% if (context['bin'+str(i+1)] == 'Unordered'):
<%
pyymap.append(pypos)
pypos += 1
%>
% for j in range(0,context['ncats'+y[i]]-1):
                tmp_ymiss.push_back(misu${context['ncats'+y[i]]}_${y[i]}_${j});
<% pyyind.append(i)%>
% endfor
% endif
% endfor
                RectMatrix mat_ymiss(tmp_ymiss, length(${y[0]}));

                std::vector<double*> tmp_u;

<% uind = 0 %>
% for i in range(0, len(y)):
% if (context['bin'+str(i+1)] == 'Unordered'):
% for m in range(0,context['ncats'+y[i]]-1):
% for j in range(0, len(context['xr'+str(i+1)+'_'+str(m)])):
                tmp_u.push_back(u${j}_${uind});
% endfor
<% uind += 1 %>
% endfor
% else:
% for j in range(0, len(context['xr'+str(i+1)])):
                tmp_u.push_back(u${j}_${uind});
% endfor
<% uind += 1 %>
% endif
% endfor

                RectMatrix mat_u(tmp_u, length(u0_0));

                SymMatrix db(${lenbeta});
                for (int i = 0; i < length(${y[0]}); i++) {
                    for (int j1 = 0; j1 < ${lenbeta}; j1++) {
                        for (int j2 = 0; j2 < j1+1; j2++) {
                            db(j1, j2) += mat_x(i, j1) * mat_d_e(betaind[j1], betaind[j2]) * mat_x(i, j2);
                        }
                    }
                }

                RectMatrix xtdy(${lenbeta}, 1);
                for (int i = 0; i < length(${y[0]}); i++) {
                    for (int j1 = 0; j1 < ${lenbeta}; j1++) {
                            
<% count = 0 %>
<% ycount = 0 %>
% for j2 in range(0, len(y)):
% if (context['bin'+str(j2+1)] == 'Unordered'):
% for m in range(0,context['ncats'+y[j2]]-1):
                            double etemp${ycount} = mat_ymiss(i, ${ycount});
% for k in range(0, len(context['xr'+str(j2+1)+'_'+str(m)])):
                            etemp${ycount} -= mat_xr(i, ${count}) * mat_u(${L2ID}[i], ${count});
<% count += 1 %>
% endfor
                            xtdy(j1, 0) += mat_x(i, j1) * mat_d_e(betaind[j1], ${ycount}) * etemp${ycount}; 
<% ycount += 1 %>
% endfor
% else:
                           double etemp${ycount} = mat_ymiss(i, ${ycount});
% for k in range(0, len(context['xr'+str(j2+1)])):
                           etemp${ycount} -= mat_xr(i, ${count}) * mat_u(${L2ID}[i], ${count});
<% count += 1 %>
% endfor
                           xtdy(j1, 0) += mat_x(i, j1) * mat_d_e(betaind[j1], ${ycount}) * etemp${ycount}; 
<% ycount += 1 %>
% endif
% endfor
                    }
                }

                fail = false;
                // int err = matrix_sym_invinplace(db);
                // RectMatrix betahat = matrix_mult(db, xtdy);
                // mat_beta = dmultnormal(betahat, db, fail);

                RectMatrix betahat = matrix_solve(db, xtdy);
                mat_beta = dmultnormalinv(betahat, db, fail);


                std::vector<int> betamap(${n});
                int pos = 0;
<% count = 0 %>
% for i in range(0, len(y)):
% if (context['bin'+str(i+1)] == 'Unordered'):
% for m in range(0,context['ncats'+y[i]]-1):
                betamap[${count}] = pos;
                pos += mat_xf${i + 1}_${m}.cols();
<% count +=1 %>
% endfor
% else:
                betamap[${count}] = pos;
                pos += mat_xf${i + 1}.cols();
<% count +=1 %>
% endif

% endfor

<% count = 0 %>
                std::vector<int> umap(${n});
                int upos = 0;
% for i in range(0, len(y)):
% if (context['bin'+str(i+1)] == 'Unordered'):
% for m in range(0,context['ncats'+y[i]]-1):
                umap[${count}] = upos;
                upos += mat_xr${i + 1}_${m}.cols();
<% count +=1 %>
% endfor
% else:
                umap[${count}] = upos;
                upos += mat_xr${i + 1}.cols();
<% count +=1 %>
% endif
% endfor



                std::vector<int> uind(${lenu});
<% ucount = 0 %>
<% count = 0 %>
% for i in range(0, len(y)):
% if (context['bin'+str(i+1)] == 'Unordered'):
% for m in range(0,context['ncats'+y[i]]-1):
% for j in range(0, len(context['xr' +str(i+1)+'_'+str(m)])):
                uind[${ucount}] = ${count};
<% ucount += 1 %>
% endfor
<% count +=1 %>
% endfor
% else:
% for j in range(0, len(context['xr' +str(i+1)])):
                uind[${ucount}] = ${count};
<% ucount += 1 %>
% endfor
<% count +=1 %>
% endif
% endfor


                for (int i = 0; i < mat_u.rows(); i++) {
                    SymMatrix duj(mat_u.cols());
                    for (int j = 0; j < mat_u.cols(); j++) {
                        for (int k = 0; k <= j; k++) {
                            duj(j, k) = mat_d_u(j, k);
                            for (int id_m = 0; id_m < idmap_${L2ID}[i].size(); id_m++) {
                                int m = idmap_${L2ID}[i][id_m];
                                duj(j, k) += mat_xr(m, j) * mat_d_e(uind[j], uind[k]) * mat_xr(m, k);
                            }
                        }
                    }

                    RectMatrix tmpuhat(mat_u.cols(), 1);
                    for (int id_k = 0; id_k < idmap_${L2ID}[i].size(); id_k++) {
                        int k = idmap_${L2ID}[i][id_k];
                        RectMatrix ymxb(${n}, 1);
<% count = 0 %>
% for j in range(0, len(y)):
                        

% if (context['bin'+str(j+1)] == 'Unordered'):
% for m in range(0,context['ncats'+y[j]]-1):
                        ymxb(${count}, 0) = mat_ymiss(k, ${count});
                        for (int m = 0; m < mat_xf${j + 1}_${m}.cols(); m++) {
                            ymxb(${count}, 0) -= mat_xf${j + 1}_${m}(k, m) * mat_beta(betamap[${count}] + m, 0);
                        }
<% count += 1 %>
% endfor
% else:
                        ymxb(${count}, 0) = mat_ymiss(k, ${count});
                        for (int m = 0; m < mat_xf${j + 1}.cols(); m++) {
                            ymxb(${count}, 0) -= mat_xf${j + 1}(k, m) * mat_beta(betamap[${count}] + m, 0);
                        }
<% count += 1 %>
% endif
% endfor
                        
                        for (int j = 0; j < mat_u.cols(); j++) {
                            for (int j1 = 0; j1 < ${n}; j1++) {
                                tmpuhat(j, 0) += mat_xr(k, j) * mat_d_e(uind[j], j1) * ymxb(j1, 0);
                            }
                        }
                    }
                    
                    // matrix_sym_invinplace(duj);
                    // RectMatrix uhat = matrix_mult(duj, tmpuhat);
                    // RectMatrix newuj = dmultnormal(uhat, duj, fail);

                    RectMatrix uhat = matrix_solve(duj, tmpuhat);
                    RectMatrix newuj = dmultnormalinv(uhat, duj, fail);

                    for (int j = 0; j < mat_u.cols(); j++) {
% if randu:
                        if (runstate != 0) {
                            mat_u(i, j) = newuj(j, 0);
                        } else {
                            mat_u(i, j) = dnorm(0, 1);
                        }
% else:
                        mat_u(i, j) = newuj(j, 0);
% endif
                    }
                }













% for i in range(0, n):

<%
mat_x_i = 'mat_xf' + str(pyyind[i] + 1)
mat_xr_i = 'mat_xr' + str(pyyind[i] + 1)
if (context['bin'+str(pyyind[i]+1)] == 'Unordered'):
    mat_x_i += '_' + str(i - pyymap[pyyind[i]])
    mat_xr_i += '_' + str(i - pyymap[pyyind[i]])
%>

                for (int j = 0; j < mat_ymiss.rows(); j++) {
                    //if (mat_y(j, ${pyyind[i]}) <= -9.999e29) {
                    {
                        double tmp = 0.0;
                        for (int k = 0; k < ${mat_x_i}.cols(); k++) {
                            tmp += mat_beta(betamap[${i}] + k, 0) * ${mat_x_i}(j, k);
                        }
                        for (int k = 0; k < ${mat_xr_i}.cols(); k++) {
                            tmp += mat_u(${L2ID}[j], umap[${i}] + k) * ${mat_xr_i}(j, k);
                        }
                        for (int k = 0; k < mat_ymiss.cols(); k++) {
                            if (k != ${i}) {
                                tmp -= mat_d_e(${i}, k) * mat_ymiss(j, k) * (1.0 / mat_d_e(${i}, ${i}));
                            }
                        }
% for m in range(0, n):
% if i != m:
<%
mat_x_m = 'mat_xf' + str(pyyind[m] + 1)
mat_xr_m = 'mat_xr' + str(pyyind[m] + 1)
if (context['bin'+str(pyyind[m]+1)] == 'Unordered'):
    mat_x_m += '_' + str(m - pyymap[pyyind[m]])
    mat_xr_m += '_' + str(m - pyymap[pyyind[m]])
%>
                        for (int k = 0; k < ${mat_xr_m}.cols(); k++) {
                            tmp += mat_d_e(${i}, ${m}) * (1.0 / mat_d_e(${i}, ${i})) * mat_u(${L2ID}[j], umap[${m}] + k) * ${mat_xr_m}(j, k);
                        }
                        for (int k = 0; k < ${mat_x_m}.cols(); k++) {
                            tmp += mat_d_e(${i}, ${m}) * (1.0 / mat_d_e(${i}, ${i})) * mat_beta(betamap[${m}] + k, 0) * ${mat_x_m}(j, k);
                        }
% endif
% endfor


% if (context['bin'+str(pyyind[i]+1)] == 'Normal'):
                    if (mat_y(j, ${pyyind[i]}) <= -9.999e29) {
                        mat_ymiss(j, ${i}) = dnorm(tmp, mat_d_e(${i}, ${i}));
                    }
% endif
% if (context['bin'+str(pyyind[i]+1)] == 'Binomial'):
                    if (mat_y(j, ${pyyind[i]}) <= -9.999e29) 
                        mat_ymiss(j, ${i}) = dnorm(tmp, mat_d_e(${i}, ${i}));
                    else if(mat_y(j, ${pyyind[i]}) <= 0) 
                        mat_ymiss(j, ${i}) = dtnormal(tmp,1/sqrt(mat_d_e(${i}, ${i})), 2, 0, 0);
                    else          
                        mat_ymiss(j, ${i}) = dtnormal(tmp,1/sqrt(mat_d_e(${i}, ${i})), 1, 0, 0);
% endif
% if (context['bin'+str(pyyind[i]+1)] == 'Ordered'):
                    if (mat_y(j, ${pyyind[i]}) <= -9.999e29) 
                        mat_ymiss(j, ${i}) = dnorm(tmp, mat_d_e(${i}, ${i}));
                    else if(mat_y(j, ${pyyind[i]}) == 0) 
                        mat_ymiss(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e(${i}, ${i})), 2, 0, 0);
                    else if(mat_y(j, ${pyyind[i]}) == ${context['ncats'+y[pyyind[i]]] - 1})
                        mat_ymiss(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e(${i}, ${i})), 1, alpha${y[i]}[ncats${y[pyyind[i]]}-2], 0);
                    else          
                        mat_ymiss(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e(${i}, ${i})), 3, alpha${y[i]}[int(mat_y(j, ${i}))-1], alpha${y[i]}[int(mat_y(j, ${i}))]);

% endif

% if (context['bin'+str(pyyind[i]+1)] == 'Unordered'):
                    if (mat_y(j, ${pyyind[i]}) <= -9.999e29) 
                        mat_ymiss(j, ${i}) = dnorm(tmp, mat_d_e(${i}, ${i}));
                    else if (mat_y(j, ${pyyind[i]}) == ${context['ncats'+y[pyyind[i]]] - 1}) 
                        mat_ymiss(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e(${i}, ${i})), 2, 0, 0);
                    else if (mat_y(j, ${pyyind[i]}) == ${i - pyymap[pyyind[i]]}) {
                        double limit = 0;
                        for (int k = 0; k < ${context['ncats'+y[pyyind[i]]] - 1}; k++) {
                            if ((k != ${i - pyymap[pyyind[i]]}) && (mat_ymiss(j, ${pyymap[pyyind[i]]} + k) > limit))
                                limit = mat_ymiss(j, ${pyymap[pyyind[i]]} + k); 
                        }
                        mat_ymiss(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e(${i}, ${i})), 1, limit, 0);
                    } else {  
                        double limit = mat_ymiss(j, ${pyymap[pyyind[i]]});
                        for (int k = 0; k < ${context['ncats'+y[pyyind[i]]] - 1}; k++) {
                            if (mat_ymiss(j, ${pyymap[pyyind[i]]} + k) > limit)
                                limit = mat_ymiss(j, ${pyymap[pyyind[i]]} + k); 
                        }
                        mat_ymiss(j, ${i}) = dtnormal(tmp, 1/sqrt(mat_d_e(${i}, ${i})), 2, 0, limit);
                    }

% endif

                    }
                }
% endfor

    }
'''

    preparedata = '''
mydata = data['datafile']

for i in range(0, len(y)):
    if (context['bin'+str(i+1)] == 'Normal'):
        context['miss'+y[i]].ncols = -1*len(mydata.variables[y[i]]['data'])
    if (context['bin'+str(i+1)] == 'Binomial'):
        context['misb'+y[i]].ncols = -1*len(mydata.variables[y[i]]['data'])
    if (context['bin'+str(i+1)] == 'Ordered'):
        context['miso'+y[i]].ncols = -1*len(mydata.variables[y[i]]['data'])
    if (context['bin'+str(i+1)] == 'Unordered'):
        for j in range(0, context['ncats'+y[i]]-1):
            context['misu'+str(context['ncats'+y[i]])+'_'+y[i]+'_'+str(j)].ncols = -1*len(mydata.variables[y[i]]['data'])
        context['origv'+str(i+1)].name.append(y[i]) 
        for var in context['origv'+str(i+1)].name:
            uniqvals = list(set(mydata.variables[var]['data'].compressed()))
            uniqvals.sort()
            uniqvals.remove(uniqvals[len(uniqvals)-1])
            for j in range(0, len(uniqvals)):
                mydata.addvariable(var + '_' + str(int(j)), data = (mydata.variables[var]['data'][:] == uniqvals[j]).astype(float))
                context['y'+str(i+1)].name.append(var + '_' + str(int(j)))
'''
