# Copyright (c) 2013, University of Bristol and University of Southampton.

from EStat.Templating import Template

class TemplateMVNlevRS(Template):
    'Fits multivariate Normal response models with one or more higher classifications (nested or crossed), allowing for random slopes, missing data, and a choice of uniform or Wishart priors.'

    __version__ = '1.0.0'

    tags = [ 'Model', 'N-Level', 'Normal', 'Multivariate response', 'Random slopes' ]
    engines = ['eStat', 'MLwiN_MCMC']

    inputs = '''
NumLevs = Integer('Number of Classifications: ')
for i in range(0, int(NumLevs)):
    selstr = 'Classification ' + str(i + 1) + ': '
    context['C' + str(i + 1)] = IDVector(selstr)
y = DataMatrix('Responses: ')

for i in range(0, len(y)):
    context['xf'+str(i+1)] = DataMatrix('Explanatory variables for response ' + y[i] + ': ', allow_cat = True)
    for j in range(0, int(NumLevs)):
        context['xr' + str(j) + '_' +str(i+1)] = DataMatrix('Explanatory variables random at ' + context['C' + str(j + 1)] + '  classification for response ' + y[i]+ '?: ', allow_cat = True)


storeresid = Boolean('Store residuals?')

lenbeta = 0
lenu = [0] * NumLevs
for i in range(0, len(y)):
    lenbeta += len(context['xf'+str(i+1)])
    context['miss' + y[i]] = ParamVector(monitor = False)
    for j in range(0, int(NumLevs)):
        lenu[j] += len(context['xr' + str(j) + '_' +str(i+1)])
        for x in range(0, len(context['xr' + str(j) + '_' +str(i+1)])):
            if storeresid:
                context['u' + str(j) + '_' + str(x) + '_' + str(i)] = ParamVector(parents=[context['C' + str(j + 1)]], as_scalar=False)
            else:
                context['u' + str(j) + '_' + str(x) + '_' + str(i)] = ParamVector(parents=[context['C' + str(j + 1)]], as_scalar=False, monitor=False)

n = len(y)
if n == 1:
    tau = ParamScalar()
    sigma = ParamScalar(modelled = False)
    for i in range(0, int(NumLevs)):
        context['tau_u' + str(i)] = ParamScalar()
        context['sigma_u' + str(i)] = ParamScalar(modelled = False)
else:
    context['omega_e'] = ParamMatrix(modelled = False, customstep = True)
    context['omega_e'].size = n
    context['d_e'] = ParamMatrix(customstep = True)
    context['d_e'].size = n
    priors = Text('Priors: ', ['Uniform', 'Wishart'])
    if priors == 'Wishart':
        R = List('R matrix: ')
        v = Integer('Degrees of Freedom:')
    for i in range(0, int(NumLevs)):
        context['omega_u' + str(i)] = ParamMatrix(modelled = False, customstep = True)
        context['omega_u' + str(i)].size = lenu[i]
        context['d_u' + str(i)] = ParamMatrix(customstep= True)
        context['d_u' + str(i)].size = lenu[i]
        context['priors' + str(i)] = Text('Priors at ' + context['C' + str(i + 1)] + ' classification: ', ['Uniform', 'Wishart'])
        if context['priors' + str(i)] == 'Wishart':
            context['R' + str(i)] = List('R matrix: ')
            context['v' + str(i)] = Integer('Degrees of Freedom:')


beta = ParamVector()
beta.ncols = lenbeta

deviance = ParamScalar(modelled = False)
'''

    model = '''
<% lenbeta = 0 %> \\

<% numlevs = int(NumLevs) %> \\
% for j in range(0, numlevs):
<% lenu = [0] * numlevs %> \\
% endfor
<% n = len(y) %> \\
model {
        for (i in 1:length(${y[0]})) {
    % if n == 1:
            miss${y[0]}[i]  ~ dnorm(mu[i], tau)
            mu[i] <- ${mmult2(context['x1'], 'beta', 'i', lenbeta)} \\
% for i in range(0, numlevs):
+ u0_0_${i}[{context['C' + str(i + 1)]}$[i]]
% endfor
<% lenbeta += len(context['xf1']) %> \\
    % else:
            dummy[i] ~ dnormal${n}a( \\
        % for i in range(0,n):
miss${y[i]}[i], \\
        % endfor
        % for i in range(0,n):
mu${i}[i], \\
        % endfor
        % for i in range(0,(((n+1)*n)/2) - 1):
d_e[${i}], \\
        % endfor
d_e[${(((n+1)*n)/2) - 1}])
        % for i in range(0, len(y)):
            mu${i}[i] <- ${mmult2(context['xf' + str(i+1)], 'beta', 'i', lenbeta)}\\
% for j in range(0, numlevs):
% for k in range(0, len(context['xr'+str(j) + '_' +str(i+1)])):
+ u${j}_${k}_${i}[${context['C' + str(j + 1)]}[i]] * ${context['xr'+str(j) + '_' +str(i+1)][k]}[i]\\
% endfor
<% lenu[j] += len(context['xr' + str(j) + '_' + str(i+1)]) %>\\
% endfor
<% lenbeta += len(context['xf' + str(i+1)]) %> 
        % endfor
        % for i in range(0,n):
        miss${y[i]}[i] ~ dflat()
        % endfor
    % endif
        }

% for j in range(0, numlevs): 
    for (i${j + 1} in 1:length(u${j}_0_0)) {     
        %if len(y) == 1 :
           u${j}_0_0[i] ~ dnorm(0, tau_u)
    % else:
        dummy_u${j}[i${j + 1}] ~ dnormal${lenu[j]}a( \\
        % for i in range(0,n):
% for k in range(0, len(context['xr'+ str(j) + '_' +str(i+1)])):
u${j}_${k}_${i}[i${j + 1}], \\
% endfor
        % endfor
        % for i in range(0,lenu[j]):
0, \\
        % endfor
        % for i in range(0,(((lenu[j]+1)*lenu[j])/2) - 1):
d_u${j}[${i}], \\
        % endfor
d_u${j}[${(((lenu[j]+1)*lenu[j])/2) - 1}])
          dum_u${j}[i${j + 1}] ~ ddummy(dummy_u${j}[i${j + 1}])

        % for i in range(0,n):
        % for k in range(0, len(context['xr'+str(j) + '_'+str(i+1)])):
        u${j}_${k}_${i}[i${j + 1}] ~ dflat()
        %endfor
        % endfor

    % endif
    }
% endfor
    # Priors
    
    % if n == 1:
    tau ~ dgamma(0.001, 0.001)
    sigma <- 1 / sqrt(tau)
    % for i in range(0, int(NumLevs)):
        tau_u${i} ~ dgamma(0.001, 0.001)
        sigma_u${i} <- 1 / sqrt(tau_u${i})
    % endfor
    % endif

    % for i in range(0, lenbeta):
    beta_${i} ~ dflat()
    % endfor   
}       
'''        

    latex = r'''
<% lenbeta = 0 %> 
<% numlevs = int(NumLevs) %>
<% n = len(y) %>
\begin{aligned} 
\left(
\begin{array}{l}
% for r in range(0, len(y)):
\mbox{${y[r]}}_i 
% if r != len(y) -1:
\\ 
% endif
% endfor
\end{array}
\right) & \sim \mbox{N}
\left[ \left(
\begin{array}{l}
% for r in range(0, len(y)):
\mu_{${r}i}
% if r != len(y) -1:
\\ 
% endif
% endfor
\end{array}
\right), \Omega_e \right] \\ 
% for r in range(0, len(y)):
\mu_{${r}i} & = ${mmulttex2(context['xf' + str(r+1)], r'\beta', 'i', lenbeta)}
% for i in range(0, numlevs):
% for j in range(0, len(context['xr' + str(i) + '_' + str(r + 1)])):
+ u_{${r},${j},${context['C' + str(i + 1)]}[i]}^{(${i + 2})} \mbox{${context['xr' + str(i) + '_' + str(r + 1)][j]}}_{i}
% endfor
% endfor
\\ 
<% lenbeta += len(context['xf' + str(r+1)]) %>
% endfor

% for lev in range(0, numlevs):
\left(
\begin{array}{l}
% for r in range(0, len(y)):
% for j in range(0, len(context['xr' + str(lev) + '_' + str(r + 1)])):
u^{(${lev + 2})}_{${r}, ${j},${context['C' + str(lev + 1)]}(i)} 
% if j != len(context['xr' + str(lev) + '_' + str(r + 1)]) - 1 or r != len(y) - 1:
\\ 
% endif
% endfor
% endfor
\end{array}
\right) & \sim \mbox{N}

\left[ \left(
\begin{array}{l}
% for r in range(0, len(y)):
% for i in range(0, len(context['xr' + str(lev) + '_' + str(r + 1)])):
0
% if i != len(context['xr' + str(lev) + '_' + str(r + 1)]) -1 or r != len(y) - 1:
\\ 
% endif
% endfor
% endfor
\end{array}
\right), \Omega^{(${lev + 2})}_{u} \right] \\ 
% endfor


% for lev in range(0, numlevs):
<% lenu = 0 %>
% for i in range(0, len(y)):
<% lenu += len(context['xr' + str(lev) + '_' + str(i+1)]) %>
% endfor

% if context['priors' + str(lev)] == 'Wishart':
\Omega^{(${lev + 2})}_u & \sim \mbox{inverse Wishart}_{${lenu}}[${context['v' + str(lev)]}*R_{u${lev + 2}},${context['v' + str(lev)]}],R_{u${lev + 2}}=
\left[\begin{array}{
% for i in range(0, lenu):
l
% endfor
}
<% lenR = 0%>
% for i in range(0, lenu):
% for j in range(0, lenu):
% if j <= i:
${context['R' + str(lev)][lenR]}
<% lenR += 1 %>
% endif
% if j != lenu - 1:
&
% else:
% if i != lenu - 1:
\\ 
% endif
% endif
% endfor
% endfor
\end{array}
\right]
% endif
% if context['priors' + str(lev)] == 'Uniform':
\Omega^{(${lev + 2})}_u & \propto 1
% endif
\\ 

% endfor


% if priors == 'Wishart':
\Omega_e & \sim \mbox{inverse Wishart}_{${len(y)}}[${v}*R_e,${v}],R_e=
\left[\begin{array}{
% for i in range(0, len(y)):
l
% endfor
}
<% lenR = 0%>
% for i in range(0, len(y)):
% for j in range(0, len(y)):
% if j <= i:
${R[lenR]}
<% lenR += 1 %>
% endif
% if j != len(y) - 1:
&
% else:
% if i != len(y) - 1:
\\ 
% endif
% endif
% endfor
% endfor
\end{array}
\right]
% endif
% if priors == 'Uniform':
\Omega_e & \propto 1
% endif
\\ 

% for i in range(0, lenbeta):
\beta_${i} & \propto 1 \\ 
% endfor   

\end{aligned} 
'''

    preccode = '''
<%!
    def mmult2(names, var, index,count):
        out = ""
        first = True
        for name in names:
            if first == False:
                out += ' + '
            else:
                first = False
            out += 'double(' + name + '[' + index + ']) * ' + var + '_' + str(count)
            count += 1
        return out
    def index(n, t):
        ind = ''
        for i in range(0, n):
            ind += '[i' + str(i + 1) + ']'
        if t != '':
            ind += '[' + t + ']'
        return ind  
%>
    {
<% n = len(y) %>\\
<% numlevs = int(NumLevs) %>\\
        // Note currently using a uniform prior for variance matrix
        SymMatrix sbe(${n});
        int vwe = length(${y[0]});
<% numbeta = 0 %>\\
        for (int i = 0; i < length(${y[0]}); i++) {
% for k in range(0, n):
            double e${k} = double(miss${y[k]}[i]) - (\\
% for j in range(0, len(context['xf' + str(k+1)])):
% if j != 0:
 + \\
% endif
double(${context['xf' + str(k+1)][j]}[i]) * beta[${numbeta}] \\
<% numbeta += 1 %>\\
% endfor
)\\
% for i in range(0, numlevs):
% for j in range(0, len(context['xr'+str(i) + '_'+str(k+1)])):
- double(u${i}_${j}_${k}[int(${context['C' + str(i + 1)]}[i])]) * double(${context['xr' + str(i) + '_' + str(k+1)][j]}[i])\\
% endfor
% endfor
;
% endfor
% for i in range(0, n):
% for j in range(i, n):
            sbe(${i}, ${j}) += e${i} * e${j};
% endfor
% endfor
        }

% if priors == 'Uniform':
        if (runstate == 0) {
% for i in range(0, n):
            sbe(${i}, ${i}) += 0.0001;
% endfor
        }
% endif
% if priors == 'Wishart':
<%
import numpy
Rmat = numpy.empty([n, n])
count = 0 
for i in range(0, n):
    for j in range(0, i + 1):
        Rmat[i, j] = float(R[count]) 
        Rmat[j, i] = Rmat[i, j]
        count += 1
%>

<% count = 0 %>
% for i in range(0, n):
% for j in range(i, n):
        sbe(${i}, ${j}) += ${str(Rmat[i, j] * float(v))};
% endfor
% endfor
% endif
        matrix_sym_invinplace(sbe);

% if priors == 'Uniform':
        vwe -= ${n + 1};
% endif
% if priors == 'Wishart':
        vwe += ${v};
% endif
        bool fail = false;
        mat_d_e = dwishart(vwe, sbe, fail); 
        mat_omega_e = matrix_sym_inverse(mat_d_e, fail);


% for k in range(0, numlevs):
        std::vector<double*> tmp_ul${k};
% for i in range(0, n):
% for j in range(0, len(context['xr'+str(k) + '_' + str(i+1)])):
        tmp_ul${k}.push_back(u${k}_${j}_${i});
% endfor
% endfor
        RectMatrix mat_u${k}(tmp_ul${k}, length(u${k}_0_0));

        SymMatrix sbu${k} = mat_u${k}.T() * mat_u${k};

<% lenu = 0 %>\\
% for i in range(0, len(y)):
<% lenu += len(context['xr' + str(k) + '_' + str(i+1)]) %>\\
% endfor

% if context['priors' + str(k)] == 'Uniform':
        if (runstate == 0) {
% for i in range(0, len(y)):
            sbu${k}(${i}, ${i}) += 0.0001;
% endfor
        }
% endif
% if context['priors' + str(k)] == 'Wishart':
<%
import numpy
Rmat = numpy.empty([len(y), len(y)])
count = 0 
for i in range(0, len(y)):
    for j in range(0, i + 1):
        Rmat[i, j] = float(context['R' + str(k)][count]) 
        Rmat[j, i] = Rmat[i, j]
        count += 1
%>

<% count = 0 %>
% for i in range(0, len(y)):
% for j in range(i, len(y)):
        sbu${k}(${i}, ${j}) += ${str(Rmat[i, j] * float(context['v' + str(k)]))};
% endfor
% endfor
% endif

        matrix_sym_invinplace(sbu${k});
% if context['priors' + str(k)] == 'Uniform':
        int vwu${k} = length(u${k}_0_0) - ${len(y) + 1};
% endif
% if context['priors' + str(k)] == 'Wishart':
        int vwu${k} = length(u${k}_0_0) + ${context['v' + str(k)]};
% endif
        mat_d_u${k} = dwishart(vwu${k}, sbu${k}, fail); 
        mat_omega_u${k} = matrix_sym_inverse(mat_d_u${k}, fail);
% endfor
    }
         '''

    mlnscript = '''
<% M=len(y) %>

% for i in range(len(y)):
MVAR 1 "${y[i]}"
% endfor
% for i in range(NumLevs):
IDEN ${i + 3} "${context['C' + str(i + 1)]}"
% endfor
IDEN 2 "_id"
IDEN 1 "resp_indicator"

<%
import numpy
T=numpy.eye(len(y), dtype='int')
%>
% for j in range(0, len(y)):
RPAT ${' '.join([str(k) for k in T[j]])}
% for i in range(0,len(context['xf'+str(j+1)])):
ADDT "${context['xf'+str(j+1)][i]}"
% endfor
ADDT "_levres"
RPAT
% endfor

% for i in range(0, len(y)):
% for k in range(0, int(NumLevs)):
% for j in range(0,len(context['xr'+str(k)+'_'+str(i+1)])):
SETV ${k + 3} "${context['xr'+str(k)+'_'+str(i+1)][j]}.${str(i+1)}"
% endfor
% endfor
SETV 2 "_levres.${i + 1}"
FPAR 0 "_levres.${i + 1}"
% endfor

<%
import numpy
n = len(y)
lenbeta =0
lenu = [0] * NumLevs
for i in range(0, len(y)):
    lenbeta += len(context['xf'+str(i+1)])
    for j in range(0, int(NumLevs)):
        lenu[j] += len(context['xr' + str(j) + '_' +str(i+1)])

join = ' '.join([str(i) for i in numpy.zeros(int(lenbeta),dtype='int')])
print 'NumLevs:' + str(int(NumLevs)) 
for i in range( int(NumLevs)-1,-1,-1):
    if context['priors' + str(i)] == 'Wishart':
        join += ' 1 '+ ' '.join(context['R' + str(i)].name) + ' ' +str(context['v' + str(i)])
    else:
        join += ' 0 '

if priors == 'Wishart':
    join += ' 1 '+ ' '.join(R.name) + ' ' + str(v)
else:
    join += ' 0 '
%>
JOIN ${join} 0 c1092
PRIOR c1092
XCLA 1
'''

    rscript = '''
<% 
M=len(y)

resp='cbind('
nparams=0
for i in range(M):
    resp+=str(y[i])
    nparams+=len(context['xf'+str(i+1)])
    if i!= M-1:
        resp+=','
    else:
        resp+=')'
TT=1
for i in range(M-1,0,-1):
    if len(context['xf'+str(i+1)])!=len(context['xf'+str(i)]):
        TT=0
    else:
        aa=context['xf'+str(i+1)].name
        aa.sort()
        bb=context['xf'+str(i)].name
        bb.sort()
        for j in range(len(aa)):
            if aa[j]!=bb[j]:
                TT=0
SS=1


for i in range(M-1,0,-1):
    for j in range(0, int(NumLevs)):
        if len(context['xr' + str(j) + '_' +str(i+1)])!=len(context['xr' + str(j) + '_' +str(i)]):
            SS=0
        else:
            aa=context['xr' + str(j) + '_' +str(i+1)].name
            aa.sort()
            bb=context['xr' + str(j) + '_' +str(i)].name
            bb.sort()
            for k in range(0,len(aa)):
                if aa[k]!=bb[k]:
                    SS=0

lenu = [0] * int(NumLevs)
for j in range(0, int(NumLevs)):
    lenu[j]=len(context['xr' + str(j) + '_1'])

%>

% if Rpackage == 'MCMCglmm':
% if TT!=0 and SS!=0:

formula <- ${resp} ~ trait:(${Rmmult(xf1)}) - 1

##define the prior...use non-informative prior #fixed effect N(0,10^6), variance~IW(1,0.002)~IG(0.001000,0.001000)
#random effect Uniform: G1~IW(vw${i+1},diag(M)*.0001); Wishart: Rmat${i+1}
#Residual Uniform: G1~IW(vw,diag(M)*.0001); Wishart: Rmat
% for i in range(0, int(NumLevs)):
% if context['priors' + str(i)] == 'Uniform':
vw${i+1} = ${M*lenu[i]  + 1};
% endif
% if context['priors' + str(i)] == 'Wishart':
vw${i+1} =  ${context['v' + str(i)]};
Rmat${i+1} = matrix(0,${M*lenu[i] },${M*lenu[i] })
Rvec${i+1}=rep(0,${len(context['R' + str(i)].name)})
% for j in range(0,len(context['R' + str(i)].name)):
Rvec${i+1}[${j+1}]=as.numeric(${float(context['R' + str(i)].name[j])})
% endfor
count = 0 
for (m in  0:${M*lenu[i] -1}){
    for (n in 0:m ){
        Rmat${i+1}[m+1, n+1] = as.numeric(Rvec${i+1}[count+1])
        Rmat${i+1}[n+1, m+1] = Rmat${i+1}[m+1, n+1]
        count = count+ 1
    }
}
% endif
% endfor

% if priors.name == 'Uniform':
vw = ${M + 1};
% endif
% if priors.name == 'Wishart':
vw = ${v.name};
Rmat = matrix(0,${M},${M})
Rvec=rep(0,${len(R.name)})
% for i in range(0,len(R.name)):
Rvec[${i+1}]=as.numeric(${float(R.name[i])})
% endfor
count = 0 
for (j in  0:${M-1}){
    for (k in 0:j ){
        Rmat[j+1, k+1] = as.numeric(Rvec[count+1])
        Rmat[k+1, j+1] = Rmat[j+1, k+1]
        count = count+ 1
    }
}
% endif
% if priors.name == 'Uniform':
myprior<-list(B=list(V=diag(${nparams})*10^6,mu=rep(0,${nparams})),G=list(
% for i in range(0,int(NumLevs)):
G${i+1}=list(V=diag(${M*lenu[i] }),nu=vw${i+1})
%if i!=int(NumLevs)-1:
,
%else:
)
% endif
% endfor
,R=list(V=diag(${M}),nu=vw))
% endif
% if priors.name == 'Wishart':
myprior<-list(B=list(V=diag(${nparams})*10^6,mu=rep(0,${nparams})),G=list(
% for i in range(0,int(NumLevs)):
G${i+1}=list(V=Rmat${i+1},nu=vw${i+1})
%if i!=int(NumLevs)-1:
,
% else:
)
% endif
% endfor
,R=list(V=Rmat,nu=vw))
% endif


myModel<-MCMCglmm(formula,random=~\\
% for i in range(0,int(NumLevs)):
us(trait:(
%for j in range(len(context['xr' + str(i) + '_1'])):
${context['xr' + str(i) + '_1'].name[j]}
%if j!=len(context['xr' + str(i) + '_1'])-1:
+
%else:
)):
%endif
%endfor
${context['C'+str(i + 1)]}\\
%if i!=int(NumLevs)-1:
+
% endif
% endfor
,rcov=~us(trait):units,family=rep("gaussian",${M}),prior=myprior,burnin=${burnin},nitt=${iterations + burnin},thin=${thinning},data=mydata,verbose=FALSE)



## summary and outputs from the model
## combined output fixed and variance into one matrix 
summary(myModel)
myModelOut<-cbind(myModel$Sol,myModel$VCV)
myModelOut<-mcmc(myModelOut)
summary(myModelOut)

##temporary plots
<% ind=divmod(nparams,3) %>

% for i in range(ind[0]):
<% svgout='DiagPlots'+str(i+1)+'.svg' %>
svg("${svgout}")
plot(myModelOut[,(1+3*${i}):(3*${i+1})])
dev.off()
% endfor

% if (ind[1]>0):
<% svgout='DiagPlots'+str(ind[0]+1)+'.svg' %>
svg("${svgout}")
plot(myModelOut[,(1+3*${ind[0]}):${nparams}])
dev.off()
% endif

## export chains
colnames(myModelOut)[which(colnames(myModelOut)=="units")]<-"sigma^2"
write.dta(data.frame(myModelOut),file="chains.dta")

% else:
<%print 'The model not implemented yet, choose another engine:'%>
% endif
% endif
'''

    preparedata = '''
mydata = data['datafile']

for i in range(0, len(y)):
    context['miss'+str(y[i])].ncols = -1*len(mydata.variables[y[i]]['data'])
'''
